﻿Imports System.IO.Ports

''' <summary>
''' The VC940 class represents the Voltcraft VC-940 multimeter. The class allows you to receive
''' data from this device.
''' 
''' Copyright (c) 2012 - Vanderhaegen Bart
'''
''' </summary>
''' <remarks>Update 16-Dec-2012</remarks>
''' 


Public Class VC940

    ' Class characteristics
    Private Const MAX_DATABUFFER_SIZE As Integer = 33
    Private Const COMMA_CHARACTER As String = "."

    ' Data buffer and pointer
    Private inBuffer(MAX_DATABUFFER_SIZE) As Byte
    Private byteCounter As Integer = 0

    ' Internal serial port object
    Private WithEvents comm As New SerialPort

    ' Defined events and delegates
    Public Event measurementReceived(ByVal measuringtype As String, ByVal unit As String, ByVal value As String, ByVal overrange As Boolean)
    Public Delegate Sub measuringReceivedDelegate(ByVal measuringtype As String, ByVal unit As String, ByVal value As String, ByVal overrange As Boolean)


    ''' <summary>
    ''' datafields is an enumeration for the 11 bytes the multimeter will transmit during each cycle
    ''' </summary>
    ''' <remarks></remarks>
    Public Enum datafields As Integer
        LCD1
        LCD2
        LCD3
        LCD4
        LCD5
        Range
        [Function]
        STATE1
        STATE2
    End Enum



#Region "Properties"
    ''' <summary>
    ''' The property comPort contains the selected COM-port
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property comPort As Integer




    ''' <summary>
    ''' The ownerForm property contains a reference to the form that holds the object instance of the
    ''' VC940 class.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property ownerForm As Form




    ''' <summary>
    ''' The property isConnected indicates if the class is currently connected to a COM-port
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property isConnected As Boolean
        Get
            Return comm.IsOpen
        End Get
    End Property
#End Region




#Region "Class Constructors"
    ''' <summary>
    ''' Constructor().
    ''' </summary>
    ''' <param name="comPort">The desired COM-port number</param>
    ''' <remarks></remarks>
    Public Sub New(comPort As Integer)
        Me.comPort = comPort
    End Sub





    ''' <summary>
    ''' Constructor()
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
    End Sub
#End Region




#Region "Event Generation"
    ''' <summary>
    ''' launchEvent will make sure the measurementReceived event is launched. The function will first
    ''' check if the ownerForm property is set. If this is the case it will invoke a delegate on the
    ''' ownerForm. Otherwise a normal event is raised.
    '''
    ''' </summary>
    ''' <param name="measuringType">Type of measurement</param>
    ''' <param name="unit">Unit of measurment</param>
    ''' <param name="value">Value of measurement</param>
    ''' <param name="overrange">Indicate if the class is in overrange</param>
    ''' <remarks></remarks>
    Public Sub launchEvent(ByVal measuringType As String, ByVal unit As String, ByVal value As String, ByVal overrange As Boolean)

        If ownerForm Is Nothing Then
            ' owner form is unknown: using delegate is not possible
            RaiseEvent measurementReceived(measuringType, unit, value, overrange)
        Else
            ' use a delegate
            If ownerForm.InvokeRequired Then
                Dim dg As New measuringReceivedDelegate(AddressOf launchEvent)
                ownerForm.Invoke(dg, New Object() {measuringType, unit, value, overrange})
            Else
                RaiseEvent measurementReceived(measuringType, unit, value, overrange)
            End If
        End If
    End Sub
#End Region







#Region "Connecting & Data receiving"
    ''' <summary>
    ''' The connect() method will establish a connection with the multimeter. The function will
    ''' return true if the connection attempt succeeds. Otherwise it will return false.
    ''' </summary>
    ''' <returns>A boolean to indicate if the connection was established or not.</returns>
    ''' <remarks></remarks>
    Public Function connect() As Boolean
        Try
            With comm
                ' Required to power the IR-cable
                .RtsEnable = False
                .DtrEnable = True

                ' Defined by Voltcraft
                .BaudRate = 2400
                .Parity = IO.Ports.Parity.None
                .DataBits = 8
                .StopBits = IO.Ports.StopBits.One

                ' Your Port
                .PortName = String.Format("COM{0}", comPort)

                ' Try to open the port
                .Open()
            End With

            Return True

        Catch
            Return False
        End Try

    End Function




    ''' <summary>
    ''' disconnect() will close the communication with the multimeter
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub disconnect()
        comm.Close()
    End Sub




    ''' <summary>
    ''' serial_DataReceived is an event which is launched when some new data is received.
    ''' If valid data was detected, the doSomeProcessing() sub is called.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub serial_DataReceived(sender As Object, e As System.IO.Ports.SerialDataReceivedEventArgs) Handles comm.DataReceived

        Static level2 As Boolean = False

        Try
            ' we add the data to the inbuffer
            Dim numberOfBytes As Integer = comm.BytesToRead

            ' here we will read the bytes one by one
            For i As Integer = 0 To numberOfBytes
                Dim readByte As Byte

                readByte = comm.ReadByte()
                inBuffer(byteCounter) = readByte

                byteCounter = byteCounter + 1

                If byteCounter >= inBuffer.Length Then
                    ' prevent overflow in the buffer
                    byteCounter = 0
                End If


                If readByte = &HD Then
                    ' data termination identifier 1
                    level2 = True

                ElseIf readByte = &H8A Then
                    ' data termination identifier 2
                    If level2 = True Then
                        ' previous character was data termination identifier 1
                        level2 = False
                        If byteCounter = 11 Then
                            Dim myData(10) As Byte
                            Array.Copy(inBuffer, myData, 11)
                            doSomeProcessing(myData)
                            byteCounter = 0
                        Else
                            byteCounter = 0
                        End If

                    Else
                        ' level 2 without level 1. Do nothing 
                    End If

                Else
                    level2 = False
                End If
            Next


        Catch ex As Exception
            Debug.Print("an error occured: " & ex.Message)
        End Try
    End Sub

#End Region




#Region "Data processing functions"


    ''' <summary>
    ''' getState1() will analyze the STATE1 packet, and return the actual state.
    ''' </summary>
    ''' <param name="state1"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function getState1(ByVal state1 As Byte) As String
        ' Only keep the second nibble
        state1 = getSecondNibble(state1)

        Select Case state1
            Case 0
                Return "OFF"
            Case 1
                Return "AC"
            Case 2
                Return "DC"
            Case 3
                Return "AC+DC"
            Case Else
                ' Should not happen
                Return ""
        End Select
    End Function



    ''' <summary>
    ''' selectValueFromList() is a helper function to select a certain item from a list of items.
    ''' </summary>
    ''' <param name="condition">The index of the item to retrieve (zero based)</param>
    ''' <param name="start">The index of the first item. It allows you to use an offset</param>
    ''' <param name="values">The list with possible items</param>
    ''' <returns>The selected object</returns>
    ''' <remarks></remarks>
    Private Function selectValueFromList(ByVal condition As Integer,
                                         ByVal start As Integer,
                                         ByVal ParamArray values() As Object) As Object
        Return values(condition - start)
    End Function




    ''' <summary>
    ''' processUnits() will do the actual processing of all data inside the multimeter.
    ''' </summary>
    ''' <param name="function"></param>
    ''' <param name="range"></param>
    ''' <param name="commaPos"></param>
    ''' <param name="unit"></param>
    ''' <param name="numOfDigits"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function processUnits(ByRef [function] As Byte,
                                  ByRef range As Byte,
                                  ByRef commaPos As Integer,
                                  ByRef unit As String,
                                  ByRef numOfDigits As Integer) As String


        Dim functionType As Integer = getSecondNibble([function])
        Dim functionSpec As Integer = getSecondNibble(range)

        Select Case functionType
            Case 0  ' AC mV
                commaPos = 3
                numOfDigits = 5
                unit = "mV"
                Return "AC mV"

            Case 1  ' DC V
                commaPos = selectValueFromList(functionSpec, 1, 1, 2, 3, 4)
                numOfDigits = 5
                unit = "V"
                Return "DC V"

            Case 2  ' AC V
                commaPos = selectValueFromList(functionSpec, 1, 1, 2, 3, 4)
                numOfDigits = 5
                unit = "V"
                Return "AC V"

            Case 3  ' DC mV
                commaPos = 3
                numOfDigits = 5
                unit = "mV"
                Return "DC mV"

            Case 4  ' Ohm
                commaPos = selectValueFromList(functionSpec, 1, 3, 1, 2, 3, 1, 2)
                numOfDigits = 4
                unit = selectValueFromList(functionSpec, 1, "Ω", "kΩ", "kΩ", "kΩ", "MΩ", "MΩ")
                Return "Ohm"

            Case 5  ' C
                commaPos = selectValueFromList(functionSpec, 1, 2, 3, 1, 2, 3, 1, 2)
                numOfDigits = 5
                unit = selectValueFromList(functionSpec, 1, "nF", "nF", "µF", "µF", "µF", "mF", "mF")
                Return "C"

            Case 6  ' °C
                commaPos = 4
                numOfDigits = 5
                unit = "°C"
                Return "°C"

            Case 7  ' µA
                commaPos = selectValueFromList(functionSpec, 0, 3, 4)
                numOfDigits = 5
                unit = "µA"
                Return "µA"

            Case 8  ' mA
                commaPos = selectValueFromList(functionSpec, 0, 2, 3)
                numOfDigits = 5
                unit = "mA"
                Return "mA"

            Case 9  ' 10 A
                commaPos = 2
                numOfDigits = 5
                unit = "A"
                Return "10 A"

            Case 10 ' Fm
                Return "Fm"

            Case 11 ' Diode
                numOfDigits = 5
                Return "Diode"

            Case 12 ' Hz
                commaPos = selectValueFromList(functionSpec, 0, 2, 3, 1, 2, 3, 1, 2, 3)
                unit = selectValueFromList(functionSpec, 0, "Hz", "Hz", "kHz", "kHz", "kHz", "MHz", "MHz", "MHz")
                numOfDigits = 5
                Return "Hz"

            Case 13 ' °F
                commaPos = 4
                numOfDigits = 5
                unit = "°F"
                Return "°F"

            Case 14 ' W
                commaPos = 4
                numOfDigits = 5
                unit = "W"
                Return "Watt"

            Case 15 ' % (4-20 mA)
                commaPos = 4
                numOfDigits = 3
                unit = "%"
                Return "% (4-20 mA)"

        End Select

        Return "?"
    End Function





    ''' <summary>
    ''' getState2() will analyze the data in the STATE2 byte
    ''' </summary>
    ''' <param name="state2"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function getState2(ByVal state2 As Byte) As String
        If (state2 And 7) = 0 Then
            Return "OFF"
        End If

        If (state2 And 1) = 1 Then
            Return "AUTO"
        End If

        If (state2 And 2) = 2 Then
            Return "MANUAL"
        End If

        ' Should not happen
        Return ""
    End Function



    ''' <summary>
    ''' insertSeparator will insert a comma (decimal point) at the given position
    ''' </summary>
    ''' <param name="lcdText"></param>
    ''' <param name="position"></param>
    ''' <remarks></remarks>
    Private Sub insertSeparator(ByRef lcdText As String, ByVal position As Integer)
        If position > 0 And position < lcdText.Length Then
            lcdText = lcdText.Insert(position, COMMA_CHARACTER)
        End If
    End Sub




    ''' <summary>
    ''' processSign() will apply the negative sign, if applicable.
    ''' </summary>
    ''' <param name="lcdText">lcdText contains the text from the LCD</param>
    ''' <param name="state2">state2 is the data byte to analyze</param>
    ''' <remarks></remarks>
    Private Sub processSign(ByRef lcdText As String, ByVal state2 As Byte)
        'TODO: mention this derives from the info in the manual!
        If (state2 And 128) = 128 Then
            lcdText = String.Format("-{0}", lcdText)
        End If
    End Sub




    ''' <summary>
    ''' doSomeProcessing() will perform the actual decoding of the received data.
    ''' </summary>
    ''' <param name="data">an array containing the required information bytes</param>
    ''' <remarks></remarks>
    Private Sub doSomeProcessing(ByVal data As Byte())

        ' Construct the text from the LCD
        Dim LCD As String =
            getSecondNibble(data(datafields.LCD1)) &
            getSecondNibble(data(datafields.LCD2)) &
            getSecondNibble(data(datafields.LCD3)) &
            getSecondNibble(data(datafields.LCD4)) &
            getSecondNibble(data(datafields.LCD5))


        ' Apply the settings from the multimeter
        Dim commaPos As Integer = 0
        Dim unit As String = ""
        Dim mode As String
        Dim numOfDigits As Integer
        mode = processUnits(data(datafields.Function),
                            data(datafields.Range),
                            commaPos,
                            unit,
                            numOfDigits)


        If LCD.Length > 5 Then
            ' Raise an event so the owner class can catch the result
            launchEvent(mode, unit, "", True)

            RaiseEvent measurementReceived(mode, unit, "", True)
        Else
            ' Insert the decimal separator at the right place 
            insertSeparator(LCD, commaPos)

            ' Process the negative sign
            processSign(LCD, data(datafields.STATE2))
            launchEvent(mode, unit, LCD, False)

            RaiseEvent measurementReceived(mode, unit, LCD, False)
        End If

    End Sub





    ''' <summary>
    ''' getSecondNubble() will get the second nibble of a byte (it will clear the first nibble)
    ''' </summary>
    ''' <param name="byteValue"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function getSecondNibble(ByVal byteValue As Byte) As Byte
        Return byteValue And &HF
    End Function

#End Region




#Region "Tools & Aider Functions"
    ''' <summary>
    ''' ByteArrayToString will show the contents of a byte array as a string. It is used for debugging.
    ''' </summary>
    ''' <param name="byteArray"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function ByteArrayToString(ByVal byteArray() As Byte) As String
        Return BitConverter.ToString(byteArray)
    End Function
#End Region


End Class
